import pytest
from json import dumps
from pyramid import testing
from pyramid.httpexceptions import HTTPBadRequest, HTTPNotFound, HTTPConflict
from clmcservice.models import ServiceFunctionEndpoint
from clmcservice.whoamiapi.views import WhoamiAPI


class TestWhoamiAPI(object):
    """
    A pytest-implementation test for the WHOAMI API endpoints
    """

    @pytest.fixture(autouse=True)
    def app_config(self):
        """
        A fixture to implement setUp/tearDown functionality for all tests by initializing configuration structure for the web service and db connection
        """

        self.registry = testing.setUp()

        yield

        testing.tearDown()
        ServiceFunctionEndpoint.delete_all()  # clear the instances of the model in the test database

    def test_get_all(self):
        """
        Tests the GET all method of the WHOAMI API - returns a list of all service function endpoint configurations from the database.
        """

        request = testing.DummyRequest()
        response = WhoamiAPI(request).get_all()
        assert response == [], "Initially there mustn't be any service function endpoint configurations in the database."

        sf_e = ServiceFunctionEndpoint(location="DC1", sfc="sfc1", sfc_i="sfc_i1", sf="sf1", sf_i="sf_i1", sf_endpoint="sf_endpoint1", sr="sr1")
        expected_response_data = [sf_e.json]
        ServiceFunctionEndpoint.add(sf_e)  # adds the new instance of the model to the database

        request = testing.DummyRequest()
        response = WhoamiAPI(request).get_all()
        assert response == expected_response_data, "Incorrect response data with 1 service function endpoint configuration."

        sf_e = ServiceFunctionEndpoint(location="DC2", sfc="sfc2", sfc_i="sfc_i2", sf="sf2", sf_i="sf_i2", sf_endpoint="sf_endpoint2", sr="sr2")
        expected_response_data.append(sf_e.json)
        ServiceFunctionEndpoint.add(sf_e)
        sf_e = ServiceFunctionEndpoint(location="DC3", sfc="sfc3", sfc_i="sfc_i3", sf="sf3", sf_i="sf_i3", sf_endpoint="sf_endpoint3", sr="sr3")
        expected_response_data.append(sf_e.json)
        ServiceFunctionEndpoint.add(sf_e)

        request = testing.DummyRequest()
        response = WhoamiAPI(request).get_all()
        assert response == expected_response_data, "Incorrect response data with more than 1 service function endpoint configurations."

    def test_get_one(self):
        """
        Tests the GET one method of the WHOAMI API - returns an instance of a service function endpoint configuration from the database.
        """

        request = testing.DummyRequest()
        response = WhoamiAPI(request).get_all()
        assert response == [], "Initially there mustn't be any service function endpoint configurations in the database."

        self._validation_of_url_parameters_test("get_one")

        sf_e = ServiceFunctionEndpoint(location="DC1", sfc="sfc1", sfc_i="sfc_i1", sf="sf1", sf_i="sf_i1", sf_endpoint="sf_endpoint1", sr="sr1")
        expected_response_data = sf_e.json
        ServiceFunctionEndpoint.add(sf_e)  # adds the new instance of the model to the database

        request = testing.DummyRequest()
        request.params["sf_endpoint"] = "sf_endpoint1"
        request.params["sf_i"] = "sf_i1"
        request.params["sr"] = "sr1"
        response = WhoamiAPI(request).get_one()
        assert response == expected_response_data, "Invalid data returned in the response of GET instance"

        request = testing.DummyRequest()
        request.params["sf_endpoint"] = "sf_endpoint2"
        request.params["sf_i"] = "sf_i2"
        request.params["sr"] = "sr2"
        error_raised = False
        try:
            WhoamiAPI(request).get_one()
        except HTTPNotFound:
            error_raised = True
        assert error_raised, "Not found error must be raised in case of a non existing service function endpoint"

    def test_post(self):
        """
        Tests the POST method of the WHOAMI API - creates an instance of a service function endpoint configuration in the database.
        """

        request = testing.DummyRequest()
        response = WhoamiAPI(request).get_all()
        assert response == [], "Initially there mustn't be any service function endpoint configurations in the database."

        resource = dict(location="DC1", sfc="sfc1", sfc_i="sfc_i1", sf="sf1", sf_i="sf_i1", sf_endpoint="sf_endpoint1", sr="sr1")
        json_data = dumps(resource)
        request = testing.DummyRequest()
        request.body = json_data.encode(request.charset)
        response = WhoamiAPI(request).post()
        assert response == resource, "POST request must return the created resource"
        assert ServiceFunctionEndpoint.exists("sf_i1", "sf_endpoint1", "sr1"), "POST request must have created the resource"

        resource["location"] = "DC2"
        json_data = dumps(resource)
        request = testing.DummyRequest()
        request.body = json_data.encode(request.charset)
        error_raised = False
        try:
            WhoamiAPI(request).post()
        except HTTPConflict:
            error_raised = True
        assert error_raised, "An error must be raised when trying to create a resource which breaks the unique constraint"

    @pytest.mark.parametrize("body, valid", [
        ('{"location": "DC1", "sfc": "sfc1", "sfc_i": "sfc_i1", "sf": "sf1", "sf_i": "sf_i1", "sf_endpoint": "sf_endpoint1", "sr": "sr1"}', True),
        ('{"location": "DC2", "sfc": "sfc2", "sfc_i": "sfc_i2", "sf": "sf2", "sf_i": "sf_i2", "sf_endpoint": "sf_endpoint2", "sr": "sr2"}', True),
        ('{}', False),
        ('{"location": "DC1", "sfc": "sfc1", "sfc_i": "sfc_i1", "sf": "sf1", "sf_i": "sf_i1"}', False),
        ('{"place": "DC2", "sfc": "sfc2", "sfc_i": "sfc_i2", "sf": "sf2", "sf_i": "sf_i2", "sf_endpoint": "sf_endpoint2", "sr": "sr2"}', False),
        ('{invalid json}', False),
    ])
    def test_post_body_validation(self, body, valid):
        """
        Tests the POST request validation of the body content.

        :param body: The request body to be validated
        :param valid: True if body is valid, False otherwise
        """

        request = testing.DummyRequest()
        request.body = body.encode(request.charset)
        error_raised = False
        try:
            WhoamiAPI(request).post()
        except HTTPBadRequest:
            error_raised = True
        assert error_raised == (not valid), "An error must be raised in case of an invalid request body"

    def test_put(self):
        """
        Tests the PUT method of the WHOAMI API - overwrites an instance of a service function endpoint configuration from the database.
        """

        request = testing.DummyRequest()
        response = WhoamiAPI(request).get_all()
        assert response == [], "Initially there mustn't be any service function endpoint configurations in the database."

        self._validation_of_url_parameters_test("put")

        resource = dict(location="location1", sfc="sfc1", sfc_i="sfc_i1", sf="sf1", sf_i="sf_i1", sf_endpoint="sf_endpoint1", sr="sr1")
        body = dumps(resource)
        request = testing.DummyRequest()
        request.params["sf_endpoint"] = "sf_endpoint1"
        request.params["sf_i"] = "sf_i1"
        request.params["sr"] = "sr1"
        request.body = body.encode(request.charset)
        error_raised = False
        try:
            WhoamiAPI(request).put()
        except HTTPNotFound:
            error_raised = True
        assert error_raised, "Not found error must be raised in case of a non existing service function endpoint"

        sf_e = ServiceFunctionEndpoint(location="DC1", sfc="sfc1", sfc_i="sfc_i1", sf="sf1", sf_i="sf_i1", sf_endpoint="sf_endpoint1", sr="sr1")
        ServiceFunctionEndpoint.add(sf_e)  # adds the new instance of the model to the database

        resource = dict(location="location1", sfc="sfc1", sfc_i="sfc_i1", sf="sf1", sf_i="sf_i1", sf_endpoint="sf_endpoint1", sr="sr1")
        body = dumps(resource)
        request = testing.DummyRequest()
        request.params["sf_endpoint"] = "sf_endpoint1"
        request.params["sf_i"] = "sf_i1"
        request.params["sr"] = "sr1"
        request.body = body.encode(request.charset)
        response = WhoamiAPI(request).put()
        assert response == resource, "PUT request must return the updated resource"
        assert ServiceFunctionEndpoint.get("sf_i1", "sf_endpoint1", "sr1").json["location"] == "location1"

        resource = dict(location="DC1", sfc="sfc1", sfc_i="sfc_i1", sf="sf1", sf_i="sf_i2", sf_endpoint="sf_endpoint2", sr="sr2")
        body = dumps(resource)
        request = testing.DummyRequest()
        request.params["sf_endpoint"] = "sf_endpoint1"
        request.params["sf_i"] = "sf_i1"
        request.params["sr"] = "sr1"
        request.body = body.encode(request.charset)
        response = WhoamiAPI(request).put()
        assert response == resource, "PUT request must return the updated resource"
        assert not ServiceFunctionEndpoint.exists("sf_i1", "sf_endpoint1", "sr1"), "Resource has not been updated"
        assert ServiceFunctionEndpoint.exists("sf_i2", "sf_endpoint2", "sr2"), "Resource has not been updated"

        sf_e = ServiceFunctionEndpoint(location="DC1", sfc="sfc1", sfc_i="sfc_i1", sf="sf1", sf_i="sf_i1", sf_endpoint="sf_endpoint1", sr="sr1")
        ServiceFunctionEndpoint.add(sf_e)  # adds the new instance of the model to the database

        resource = dict(location="DC1", sfc="sfc1", sfc_i="sfc_i1", sf="sf1", sf_i="sf_i2", sf_endpoint="sf_endpoint2", sr="sr2")
        body = dumps(resource)
        request = testing.DummyRequest()
        request.params["sf_endpoint"] = "sf_endpoint1"
        request.params["sf_i"] = "sf_i1"
        request.params["sr"] = "sr1"
        request.body = body.encode(request.charset)
        error_raised = False
        try:
            WhoamiAPI(request).put()
        except HTTPConflict:
            error_raised = True
        assert error_raised, "PUT request validates unique constraint"

    def test_delete(self):
        """
        Tests the DELETE method of the WHOAMI API - deletes an instance of a service function endpoint configuration from the database.
        """

        request = testing.DummyRequest()
        response = WhoamiAPI(request).get_all()
        assert response == [], "Initially there mustn't be any service function endpoint configurations in the database."

        self._validation_of_url_parameters_test("delete")

        sf_e = ServiceFunctionEndpoint(location="DC1", sfc="sfc1", sfc_i="sfc_i1", sf="sf1", sf_i="sf_i1", sf_endpoint="sf_endpoint1", sr="sr1")
        ServiceFunctionEndpoint.add(sf_e)  # adds the new instance of the model to the database

        assert ServiceFunctionEndpoint.exists("sf_i1", "sf_endpoint1", "sr1")

        request = testing.DummyRequest()
        request.params["sf_endpoint"] = "sf_endpoint1"
        request.params["sf_i"] = "sf_i1"
        request.params["sr"] = "sr1"
        response = WhoamiAPI(request).delete()
        assert response == {}, "DELETE must return an empty body if successful"

        assert not ServiceFunctionEndpoint.exists("sf_i1", "sf_endpoint1", "sr1"), "Resource must be deleted after the delete API method has been called."

        request = testing.DummyRequest()
        request.params["sf_endpoint"] = "sf_endpoint1"
        request.params["sf_i"] = "sf_i1"
        request.params["sr"] = "sr1"
        error_raised = False
        try:
            WhoamiAPI(request).delete()
        except HTTPNotFound:
            error_raised = True
        assert error_raised, "Not found error must be raised in case of a non existing service function endpoint"

    @staticmethod
    def _validation_of_url_parameters_test(method):
        """
        Validates the way a whoami API method handles url query parameters

        :param method: the method to test
        """

        request = testing.DummyRequest()
        error_raised = False
        try:
            getattr(WhoamiAPI(request), method).__call__()
        except HTTPBadRequest:
            error_raised = True
        assert error_raised, "Error must be raised in case of no URL parameters"

        request = testing.DummyRequest()
        request.params["sf_i"] = "sf_i"
        request.params["sr"] = "sr"
        try:
            getattr(WhoamiAPI(request), method).__call__()
        except HTTPBadRequest:
            error_raised = True
        assert error_raised, "Error must be raised in case of insufficient number of arguments"

        request = testing.DummyRequest()
        request.params["sf_endp"] = "sf_endpoint"  # argument should be sf_endpoint
        request.params["sf_i"] = "sf_i"
        request.params["sr"] = "sr"
        try:
            getattr(WhoamiAPI(request), method).__call__()
        except HTTPBadRequest:
            error_raised = True
        assert error_raised, "Error must be raised in case of invalid naming of arguments"
